#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
#
# This script is licensed under GNU GPL version 2.0 or above
# (c) 2023 Antonio J. Delgado
# List IMAP messages

import sys
import os
import logging
from logging.handlers import SysLogHandler
import imaplib
import json
import click
import click_config_file

class Mailbox:
    '''Class defining an IMAP mailbox'''
    def __init__(self, name, count, children):
        self.name = name
        self.count = count
        self.children = children

    def __str__(self):
        return f"{self.name} {self.count} messages and {len(self.children)} children mailboxes"

    def to_json(self):
        '''Convert mailbox object to JSON data'''
        return json.dumps(self, default=lambda o: o.__dict__, 
            sort_keys=True, indent=4)

    def __repr__(self):
        return repr((self.name, self.count, self.children))

class ListImapMessages:
    '''Class to list IMAP messages'''
    def __init__(self, debug_level, log_file, **kwargs):
        ''' Initial function called when object is created '''
        self.config = {}
        self.config['debug_level'] = debug_level
        if log_file is None:
            log_file = os.path.join(
                os.environ.get(
                    'HOME',
                    os.environ.get(
                        'USERPROFILE',
                        os.getcwd()
                    )
                ),
                'log', 'list_imap_messages.log'
            )
        self.config['log_file'] = log_file
        self.config.update(kwargs)
        self._init_log()
        if self.config['ssl']:
            self._connect_ssl()
        else:
            self._connect()
        self._login()
        #self._select_mailbox()
        self._list_mailboxes()

    def _list_mailboxes(self):
        self._log.debug("Listing mailboxes in '%s'", self.config['mailbox'])
        result_list = self.imap.list()
        if result_list[0] == 'NO':
            self._log.error("Error listing mailboxes in %s: %s", self.config['mailbox'], result_list[1])
            self.imap.close()
            self.imap.logout()
            sys.exit(1)
        else:
            self._log.debug("Result: %s", result_list)
        mailboxes = list()
        for raw_mailbox in result_list[1]:
            self._log.debug(raw_mailbox)
            mailbox_list = raw_mailbox.decode().split(' "/" ')
            result = self._select_mailbox(mailbox_list[1])
            self._log.debug("Selecting mailbox '%s', result: %s", mailbox_list[1], result)
            # result_list = self.imap.list(mailbox_list[1])[1]
            # self._log.debug(f"Listing mailboxes in '{mailbox_list[1]}', result: {result_list}")
            # if len(result_list) > 1:
            #     children_mailboxes = result_list[1]
            # else:
            #     children_mailboxes = []
            if b'hasnochildren' in raw_mailbox:
                children_mailboxes = []
            else:
                children_mailboxes = self._get_children(mailbox_list[1], result_list[1])
            count = int(result[1][0])
            if (self.config['only_empty'] and count == 0 and len(children_mailboxes) == 0) or not self.config['only_empty']:
                mailboxes.append(
                    Mailbox(
                        name=mailbox_list[1].replace('"', ''),
                        count=count,
                        children=children_mailboxes
                        )
                    )
        self._log.info(json.dumps(sorted(mailboxes, key=lambda mailbox: mailbox.count, reverse=True), indent=2, default=vars))
        return result

    def _get_children(self, mailbox, mlist):
        children = [ mailbox ]
        for item in mlist:
            item_string = item.decode().split(' "/" ')[1].replace('"', '')
            if mailbox in item_string:
                if item_string != mailbox:
                    children.append(item_string)
        return children

    def _select_mailbox(self, mailbox):
        try:
            self._log.debug("Selecting mailbox '%s@%s:%s/%s'", self.config['username'], self.config['server'], self.config['port'], mailbox)
            result=self.imap.select(mailbox, True)
        except imaplib.IMAP4.error as error:
            self._log.error("Error selecting mailbox '%s@%s:%s/%s': %s", self.config['username'], self.config['server'], self.config['port'], mailbox, error)
            self.imap.close()
            self.imap.logout()
            sys.exit(1)
        if result[0] == 'NO':
            self._log.error("Error selecting mailbox '%s@%s:%s/%s': %s", self.config['username'], self.config['server'], self.config['port'], mailbox, result[1])
            self.imap.close()
            self.imap.logout()
            sys.exit(1)
        else:
            self._log.debug("Result: %s", result)
        return result

    def _login(self):
        try:
            self._log.debug("Login as '%s@%s:%s'...", self.config['username'], self.config['server'], self.config['port'])
            self.imap.login(self.config['username'], self.config['password'])
        #except imaplib.IMAP4.error as error:
        except Exception as error:
            self._log.error("Error login as '%s@%s:%s': %s", self.config['username'], self.config['server'], self.config['port'], error)
            self.imap.logout()
            sys.exit(1)

    def _connect_ssl(self):
        try:
            self._log.debug("Connecting to '%s:%s/SSL'...", self.config['server'], self.config['port'])
            self.imap=imaplib.IMAP4_SSL(self.config['server'], self.config['port'])
        except Exception as error:
            self._log.error("Error connecting to '%s:%s': %s", self.config['server'], self.config['port'], error)
            sys.exit(1)

    def _connect(self):
        try:
            self._log.debug("Connecting to '%s:%s/noSSL'...", self.config['server'], self.config['port'])
            self.imap=imaplib.IMAP4(self.config['server'], self.config['port'])
        except Exception as error:
            self._log.error("Error connecting to '%s:%s': %s", self.config['server'], self.config['port'], error)
            sys.exit(1)


    def _init_log(self):
        ''' Initialize log object '''
        self._log = logging.getLogger("list_imap_messages")
        self._log.setLevel(logging.DEBUG)

        sysloghandler = SysLogHandler()
        sysloghandler.setLevel(logging.DEBUG)
        self._log.addHandler(sysloghandler)

        streamhandler = logging.StreamHandler(sys.stdout)
        streamhandler.setLevel(logging.getLevelName(self.config.get("debug_level", 'INFO')))
        self._log.addHandler(streamhandler)

        if 'log_file' in self.config:
            log_file = self.config['log_file']
        else:
            home_folder = os.environ.get('HOME', os.environ.get('USERPROFILE', ''))
            log_folder = os.path.join(home_folder, "log")
            log_file = os.path.join(log_folder, "list_imap_messages.log")

        if not os.path.exists(os.path.dirname(log_file)):
            os.mkdir(os.path.dirname(log_file))

        filehandler = logging.handlers.RotatingFileHandler(log_file, maxBytes=102400000)
        # create formatter
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        filehandler.setFormatter(formatter)
        filehandler.setLevel(logging.DEBUG)
        self._log.addHandler(filehandler)
        return True

@click.command()
@click.option("--debug-level", "-d", default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        case_sensitive=False,
    ), help='Set the debug level for the standard output.')
@click.option('--log-file', '-l', help="File to store all debug messages.")
@click.option("--ssl","-s", is_flag=True, default=True, help="Use SSL connection usually port 993.")
@click.option("--server","-S", default='localhost', help="Server full name or IP.")
@click.option("--port","-p", default=993, help="Server port to connect.")
@click.option("--username","-u", required=True, help="User name to login.")
@click.option("--password","-w", required=True, help="User password to login.")
@click.option("--mailbox","-m", default='INBOX', help="Mailbox to start examination.")
@click.option("--only-empty", "-e", is_flag=True, default=False, help='List only empty folder')
@click_config_file.configuration_option()
def __main__(debug_level='INFO', log_file=None, **kwargs):
    return ListImapMessages(debug_level, log_file, **kwargs)

if __name__ == "__main__":
    __main__()

